#!/usr/bin/python3
#
# Workflow:
#
#      Data from Host                   Ids Server
#            |                              |
#       EventInfoMsg         --->    Consume, flatten,   --->           EventInfo           --->        Druid
#   [ids_data kafka topic]                filter                 [ids_event kafka topic]           [ids_event table]
#
import os
import configparser
import ipaddress
import json
import signal
import threading
import uuid
import tls_helper
import vmware.nsx.context.ids.ids_event_pb2 as ids_event_pb2
from ids_server_logging import setup_logger, get_logger
from kafka import KafkaProducer, KafkaConsumer

# Initialize logger
logger = get_logger(__name__)

#
# Kafka consumer param keys - should match the keys in the config file
#
CONFIG_FILE_PATH = os.getenv('IDS_SERVER_CONFIG_FILE',
                             "/opt/vmware/config/ids_server.cfg")

CONFIG_SECTION_KAFKA = 'Kafka'
CFG_HOST_TOPIC = 'host_topic'
CFG_DRUID_TOPIC = 'druid_topic'
CFG_KAFKA_SERVER = 'kafka_server'
CFG_KAFKA_SECURITY_PROTOCOL = 'SSL'

CONFIG_SECTION_CONSUMER = 'Consumer'
CONSUMER_TIMEOUT_MS = 'consumer_timeout_ms'

# Filter hosts json file. Contains list of hosts to include in filtering of Ids events
FILTER_HOSTS_FILE = os.getenv('FILTER_HOSTS_FILE', 'filter_hosts.json')

#
# Ids Consumer class that implements the ids event consumption logic
# The consumer runs as a thread and polls kafka topic for new ids_data messages. It processes each envelope message,
# flattens it out and extracts each ids_event message.
#


class IdsServer(threading.Thread):
    def __init__(self, ctrl, hosts):
        threading.Thread.__init__(self)
        self.ctrl = ctrl
        self.filter_hosts = hosts
        self.producer = None
        self.start()

    def run(self):
        # Read Kafka and Consumer configuration from config file
        try:
            config = read_config_params()
        except Exception as e:
            logger.error('Exception in reading config params, (%s)' % e)
            return

        consumer = None
        try:
            tls_helper.create_pem_from_pkcs12(p12_file_path=config[CONFIG_SECTION_KAFKA]["ssl_keystore_file_p12"],
                                              pem_file_path=config[CONFIG_SECTION_KAFKA]["ssl_keystore_file"])
            tls_helper.create_pem_from_pkcs12(p12_file_path=config[CONFIG_SECTION_KAFKA]["ssl_truststore_file_p12"],
                                              pem_file_path=config[CONFIG_SECTION_KAFKA]["ssl_truststore_file"])
            consumer = KafkaConsumer(config[CONFIG_SECTION_KAFKA][CFG_HOST_TOPIC],
                                     bootstrap_servers=config[CONFIG_SECTION_KAFKA][CFG_KAFKA_SERVER],
                                     security_protocol=CFG_KAFKA_SECURITY_PROTOCOL,
                                     ssl_keyfile=config[CONFIG_SECTION_KAFKA]["ssl_keystore_file"],
                                     ssl_certfile=config[CONFIG_SECTION_KAFKA]["ssl_keystore_file"],
                                     ssl_cafile=config[CONFIG_SECTION_KAFKA]["ssl_truststore_file"],
                                     consumer_timeout_ms=config[CONFIG_SECTION_CONSUMER].getint(CONSUMER_TIMEOUT_MS))
            logger.info('Created consumer for topic: %s' % config[CONFIG_SECTION_KAFKA][CFG_HOST_TOPIC])

            consumer.subscribe(config[CONFIG_SECTION_KAFKA][CFG_HOST_TOPIC])
            logger.info('Consumer subscribed to topic: %s' % config[CONFIG_SECTION_KAFKA][CFG_HOST_TOPIC])

            self.producer = KafkaProducer(bootstrap_servers=config[CONFIG_SECTION_KAFKA][CFG_KAFKA_SERVER],
                                          security_protocol=CFG_KAFKA_SECURITY_PROTOCOL,
                                          ssl_keyfile=config[CONFIG_SECTION_KAFKA]["ssl_keystore_file"],
                                          ssl_certfile=config[CONFIG_SECTION_KAFKA]["ssl_keystore_file"],
                                          ssl_cafile=config[CONFIG_SECTION_KAFKA]["ssl_truststore_file"],
                                          value_serializer=lambda v: json.dumps(v).encode('utf-8'))
            logger.info('Created producer for topic: %s' % config[CONFIG_SECTION_KAFKA][CFG_DRUID_TOPIC])

            # Start consuming on host topic
            msg_count = 0
            while not self.ctrl.is_shutdown():
                consumer.poll()
                for msg in consumer:
                    if msg is not None:
                        msg_count = msg_count + 1
                        self.process_msg(config[CONFIG_SECTION_KAFKA][CFG_DRUID_TOPIC], msg, msg_count)

            logger.info('Processed %d msgs' % msg_count)
            logger.info('Shutting down...')
        except Exception as e:
            logger.error('Exception consumer/producer/msg processing, (%s)' % e)
        finally:
            if self.producer is not None:
                self.producer.flush()
                self.producer.close()
                logger.info('Flushed and closed producer')
            if consumer is not None:
                consumer.unsubscribe()
                # Close down consumer to commit final offsets.
                consumer.close()
                logger.info('Unsubscribed and closed consumer')

    #
    # Process ids events
    #
    def process_msg(self, druid_topic, msg, id):
        logger.info('Processing msg %d' % id)
        event_info_msg = ids_event_pb2.EventInfoMsg()
        event_info_msg.ParseFromString(msg.value)

        #  1. Check if host in the msg is enabled for ids data collection
        #  2. If enabled
        #       2.1 Extract ids_event messages for filtered hosts
        #       2.2 Convert ids_event protobuf messages to json
        #       2.3 Send messages to druid kafka topic
        if self.filter_hosts.filter(uuid_to_str(event_info_msg.tn_uuid)):
            logger.info('Processing %d events from host %s' % (len(event_info_msg.event_info),
                                                               uuid_to_str(event_info_msg.tn_uuid)))
            self.convert_send_events(druid_topic, event_info_msg)
        else:
            logger.info('Skipping msg %d' % id)

    #
    # Utility methods
    #
    def convert_send_events(self, druid_topic, event_info_msg):
        bool_to_flow_direction = {True: 'inbound', False: 'outbound'}
        bool_to_similar_flow_direction = {True: 'similar', False: 'different'}

        int_to_session_direction = {0: 'SESSION_DIRECTION_INVALID',
                                    1: 'SESSION_DIRECTION_LSP_AS_SOURCE',
                                    2: 'SESSION_DIRECTION_LSP_AS_DESTINATION'}

        int_to_event_direction = {0: 'EVENT_DIRECTION_INVALID',
                                  1: 'EVENT_DIRECTION_INGRESS',
                                  2: 'EVENT_DIRECTION_EGRESS',
                                  3: 'EVENT_DIRECTION_BOTH'}

        int_to_effective_action = {0: 'EFFECTIVE_ACTION_ALERT',
                                   1: 'EFFECTIVE_ACTION_DROP',
                                   2: 'EFFECTIVE_ACTION_REJECT',
                                   3: 'EFFECTIVE_ACTION_INVALID'}

        for event_num, event in enumerate(event_info_msg.event_info, start=1):
            ids_json_event = {}
            ids_json_event['signatureId'] = event.signature_id
            ids_json_event['signatureName'] = event.signature_name
            ids_json_event['sourceIpAddress'] = ip_to_str(event.source_ip_address)
            ids_json_event['destinationIpAddress'] = ip_to_str(event.destination_ip_address)

            ids_json_event['flowDirSimilarToAttack'] = bool_to_similar_flow_direction[event.flow_dir_similar_to_attack]
            ids_json_event['flowDirectionInbound'] = bool_to_flow_direction[event.flow_direction_inbound]
            ids_json_event['sessionDirection'] = int_to_session_direction[event.session_direction]
            ids_json_event['eventDirection'] = int_to_event_direction[event.event_direction]

            ids_json_event['protocol'] = event.protocol
            ids_json_event['appProtocol'] = event.app_protocol
            ids_json_event['severity'] = event.severity
            ids_json_event['vmUuid'] = event.vm_uuid
            ids_json_event['lspUuid'] = event.lsp_uuid
            ids_json_event['userId'] = event.user_id
            ids_json_event['logicalRouterId'] = uuid_to_str(event.logical_router_id)
            ids_json_event['profileId'] = event.profile_id
            ids_json_event['effectiveAction'] = int_to_effective_action[event.effective_action]
            ids_json_event['ruleId'] = event.rule_id
            ids_json_event['vnicIndex'] = event.vnic_index

            ids_json_event['count'] = event.count
            ids_json_event['bytesToServer'] = event.bytes_toserver
            ids_json_event['bytesToClient'] = event.bytes_toclient
            ids_json_event['eventTime'] = event.event_time

            logger.info('Sending event %d: ' % event_num)
            logger.debug(ids_json_event)
            self.producer.send(druid_topic, ids_json_event)


#
# Helper class to hold list of hosts to filter (inclusion filter)
#
class FilterHosts:
    def __init__(self):
        self.filter_hosts = {}
        self.filter_hosts_lock = threading.Lock()

    # Filter hosts using a json file
    # Default behavior:
    #   If filter hosts file is not found or there is any error in json parsing then filter all hosts
    #   i.e. collect events from all hosts
    def init_filter_hosts(self, filter_hosts_json):
        with self.filter_hosts_lock:
            try:
                self.filter_hosts = json.load(open(filter_hosts_json))
                logger.info('Filtering hosts %s from file %s' % (self.filter_hosts, filter_hosts_json))
            except Exception as e:
                logger.warning('Exception (%s) while processing %s, filtering(including) all hosts' %
                               (e, FILTER_HOSTS_FILE))

    # tn_uuid is expected in string format
    # Default behavior:
    #   If hosts are not configured for filtering or filter_hosts list is empty then return true
    #   i.e. collect events from all hosts
    def filter(self, tn_uuid):
        found = False
        with self.filter_hosts_lock:
            if len(self.filter_hosts) == 0:
                found = True
            else:
                for host in self.filter_hosts:
                    if host == tn_uuid:
                        found = True
            return found


#
# Helper class to control and shutdown all threads gracefully
#
class MonitorCtrl:
    def __init__(self):
        self.stopCv = threading.Condition(threading.Lock())
        self.stop = False

    def is_shutdown(self):
        with self.stopCv:
            shutdown = self.stop
        return shutdown

    def signal_shutdown(self, signum, frame):
        logger.info('Received signal %d, shutting down...' % signum)
        with self.stopCv:
            self.stop = True
            self.stopCv.notifyAll()

    def wait_for_shutdown(self, timeout=300):
        with self.stopCv:
            if self.stop is False:
                self.stopCv.wait(timeout)


#
# Static utility functions
#
def ip_to_str(ip_msg):
    ip_str = ""
    try:
        if ip_msg.HasField('ipv4'):
            ip_str = str(ipaddress.IPv4Address(ip_msg.ipv4))
            # logger.debug('Converting ipv4 to string %s' % ip_str)
        else:
            if ip_msg.HasField('ipv6'):
                ip_str = str(ipaddress.IPv6Address(ip_msg.ipv6))
                # ip_str = socket.inet_ntop(socket.AF_INET6, ip_msg.ipv6)
                # logger.debug('Converting ipv6 to string %s' % ip_str)
    except Exception as e:
        logger.error('Exception in conversion, using empty str, (%s)' % e)
    return ip_str


def uuid_to_str(uuid_msg):
    uuid_str = ""
    try:
        uuid_int = int(uuid_msg.left << 64 | uuid_msg.right)
        uuid_str = str(uuid.UUID(int=uuid_int))
    except Exception as e:
        logger.error('Exception in conversion, using empty str, (%s)' % e)
    return uuid_str


def read_config_params():
    config = configparser.ConfigParser()
    found = config.read(CONFIG_FILE_PATH)
    if not found:
        raise Exception('read_config_params: Failed to read config file %s' % CONFIG_FILE_PATH)
    return config


#
# Entry point of this server
#
def main():
    try:
        # Configure logging
        setup_logger()
        ctrl = MonitorCtrl()
        signal.signal(signal.SIGINT, ctrl.signal_shutdown)
        signal.signal(signal.SIGTERM, ctrl.signal_shutdown)

        # Initialize FilterHosts object
        hosts = FilterHosts()
        hosts.init_filter_hosts(FILTER_HOSTS_FILE)

        # Start consumer thread
        logger.info('Starting IdsServer')
        ids_consumer = IdsServer(ctrl, hosts)

        # Wait for consumer thread
        logger.info('Waiting for IdsServer thread to finish...')
        ids_consumer.join()

        logger.info('Done IdsServer!')

    except Exception as e:
        logger.error('Exception (%s)' % e)


if __name__ == "__main__":
    main()
